package com.regexlab.toys;

import java.io.*;
import java.util.regex.*;

public class SvnDumpUp {

    public static final int M        = 1024 * 1024;
    public static final int defaultJ = 1024 * 200;
    
    /**
     * Entry point of tool
     */
    public static void main(String[] args) throws Exception {

        if(args.length < 2)
        {
            System.out.println("Usage: svndumpup <dumpfile> <parent-to-remove>");
            return;
        }
        
        BufferedInputStream in = new BufferedInputStream(new FileInputStream(args[0]));
        BufferedOutputStream out = new BufferedOutputStream(new FileOutputStream(args[0] + ".up"));
        
        byte [] buf = new byte[M];
        int lastJ = 0;
        
        int red = 0;
        while( (red = in.read(buf, lastJ, M - lastJ)) > 0 ) {
            
            String content = new String(buf, 0, lastJ + red, "iso-8859-1");
            
            UpResult result = doUp(content, args[1]);
            
            // ok part
            out.write(result.result.getBytes("iso-8859-1"));

            // left part
            byte [] restBytes = result.rest.getBytes("iso-8859-1");
            lastJ = restBytes.length;
            System.arraycopy(restBytes, 0, buf, 0, lastJ);
        }
        
        // left bytes
        out.write(buf, 0, lastJ);
        
        out.close();
        in.close();
    }
    
    private static UpResult doUp(String content, String string) {
        
        StringBuffer sb = new StringBuffer();
        int lastpos = 0;
        Matcher m = nodePath.matcher(content);
        
        while(m.find()) {
            sb.append(content.substring(lastpos, m.start()));
            
            String path = m.group(2);
            
            if(path.length() <= string.length() && string.startsWith(path)) {
                // del
                System.out.println("Del " + path);
            }
            else if(path.startsWith(string)) {
                String path1 = path.substring(string.length() + 1);
                System.out.println("Move " + path + " => " + path1);
                
                sb.append(m.group(1)).append(path1).append(doUpCopyfromPath(m.group(3), string));
            }
            else {
                // skip
                System.out.println("Skip " + path);
                sb.append(m.group(1)).append(path).append(doUpCopyfromPath(m.group(3), string));
            }
            
            lastpos = m.end();
        }
        
        UpResult result = new UpResult();
        
        result.result = sb.toString();
        result.rest   = content.substring(lastpos);
        
        if(result.rest.length() > defaultJ) {
            result.result += result.rest.substring(0, result.rest.length() - defaultJ);
            result.rest    = result.rest.substring(result.rest.length() - defaultJ);
        }
        
        return result;
    }

    private static String doUpCopyfromPath(String source, String prefix) {
        
        Matcher m = copyfromPath.matcher(source);
        
        if(m.find()) {
            String from = m.group(1);
            
            return doSpecial_1(source.substring(0, m.start(1)), from, prefix) + 
                    doMergeinfo(source.substring(m.end(1)), prefix);
        }

        return doMergeinfo(source, prefix);
        
    }

    private static String doMergeinfo(String substring, String prefix) {
        
        Matcher m = regexMergeinfo.matcher(substring);
        
        if(m.find()) {
            
            String lines = m.group(2);
            {
                StringBuffer sb = new StringBuffer();
                Matcher mm = regexLine.matcher(lines);
                int lastpos = 0;
                while(mm.find()) {
                    sb.append(lines.substring(lastpos, mm.start()));
                    
                    String line = mm.group();
                    if(line.startsWith(prefix) || line.startsWith("/" + prefix))
                        sb.append(line.substring(prefix.length() + 1));
                    else
                        sb.append(line);
                    
                    lastpos = mm.end();
                }
                
                sb.append(lines.substring(lastpos));
                
                lines = sb.toString();
            }
            
            StringBuffer mergeinfo = new StringBuffer();
            
            // before
            mergeinfo.append(substring.substring(0, m.start(1)));
            
            // length
            mergeinfo.append(lines.length() - 1);
            
            // after
            mergeinfo.append(substring.substring(m.end(1), m.start(2)));
            
            // data
            mergeinfo.append(lines);
            
            // after data
            mergeinfo.append(substring.substring(m.end(2)));
            
            return mergeinfo.toString();
        }
        
        return substring;
    }

    private static String doSpecial_1(String substring, String from, String prefix) {
        return substring + from.substring(prefix.length() + 1);
    }

    // regex
    static Pattern nodePath = Pattern.compile("(Node-path: )([^\\r\\n]+)(((?!PROPS-END|Node-path).)*(PROPS-END)?)", Pattern.DOTALL);

    static Pattern copyfromPath = Pattern.compile("Node-copyfrom-path: ([^\\r\\n]*)");
    
    static class UpResult {
        public String result;
        public String rest;
    }
    
    // regex
    static Pattern regexMergeinfo = Pattern.compile("^K\\s+\\d+\\s+\nsvn:mergeinfo\\s+\n^V\\s+(\\d+)\\s+\n((?:(?!^PROPS-END|^K\\s+\\d+).)+)", Pattern.DOTALL | Pattern.MULTILINE | Pattern.COMMENTS);
    
    static Pattern regexLine = Pattern.compile("^[^\\r\\n]*", Pattern.MULTILINE);
}
